"""
fritz.api_resources.model_version
~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

:copyright: © 2019 by Fritz Labs Incorporated
:license: MIT, see LICENSE for more details.
"""
import time
import logging

import fritz
import fritz.utils
import fritz.errors
from fritz.api_resources.fritz_object import FritzObject
from fritz.api_client import FritzClient

_logger = logging.getLogger(__name__)


class ModelVersion(FritzObject):
    """ModelVersion  """

    OBJECT_NAME = "ModelVersion"

    # pylint: disable=duplicate-code
    @classmethod
    def create(
        cls,
        api_key=None,
        project_uid=None,
        model_uid=None,
        snapshot_uid=None,
        filename=None,
        data=None,
        set_active=None,
        metadata=None,
    ):
        """Create a new ModelVersion, uploading file to Fritz.

        Args:
            api_key (Optional[str]): API Key. Optional if `fritz.configure`
                was called.
            project_uid (Optional[str]): Project UID. Optional if
                `fritz.configure` was called with project_uid.
            model_uid (str): optional Model UID.
            snapshot_uid (str): If snapshot_uid set, will attach version to
                given snapshot.
            filename (str): Name of model version filename.
            set_active (bool): If True, model will be set as the active version
               in Fritz. If it is True, any devices runninng this model will
               download the latest version.
            metadata (dict): Dictionary of metadata.

        Returns:
            Tuple[ModelVersion, ModelSnapshot, Model]
        """
        if not fritz.api_key and not api_key:
            raise fritz.errors.FritzNotInitializedError()

        if model_uid:
            url = "/model/{model_uid}/version".format(model_uid=model_uid)
        else:
            url = "/model/version"

        client = FritzClient(api_key or fritz.api_key)
        response = client.post(
            url,
            params={
                "project_uid": project_uid or fritz.project_uid,
                "snapshot_uid": snapshot_uid,
                "model_uid": model_uid,
            },
            data={"set_active": set_active, "metadata": metadata},
            files={"file": (filename, data)},
        )
        converted = {
            key: fritz.utils.convert_to_fritz_object(value)
            for key, value in response.items()
        }
        return converted["version"], converted["snapshot"], converted["model"]

    @classmethod
    def get(cls, version_uid, api_key=None):
        """Get version by version_uid

        Args:
            version_uid: Version UID
            api_key (str): Optional API Key to use.

        Returns: ModelVersion
        """
        url = "/model/version/{version_uid}".format(version_uid=version_uid)

        client = FritzClient(api_key or fritz.api_key)

        response = client.get(url)
        return fritz.utils.convert_to_fritz_object(response)

    def benchmark(self, api_key=None, wait_seconds=5, attempts=5):
        """Get model grade report, waiting if it does not yet exist.

        Args:
            api_key (str): Optional API Key
            wait_seconds (int): Number of seconds to wait between each
                request.
            attempts (int): Number of attempts to make.

        Returns: ModelGradeReport if it exists
        """

        try:
            return fritz.ModelGradeReport.get(self.uid, api_key=api_key)
        except fritz.errors.FritzError as err:
            if err.status_code != 404:
                raise err
            _logger.info("Grader report not found, trying again")
            attempts -= 1
            if attempts < 0:
                raise err

            time.sleep(wait_seconds)

            return self.benchmark(
                api_key=api_key, wait_seconds=wait_seconds, attempts=attempts
            )



import collections
import tensorflow as tf
IMAGE_FORMAT = 'jpeg'
LABEL_FORMAT = 'png'


def int64_list_feature(values):
    """Return a TF-Feature of int64_list.

    Args:
        values: A scalar or list of values.
    Returns:
        A TF-Feature.
    """
    if not isinstance(values, collections.Iterable):
        values = [values]

    return tf.train.Feature(int64_list=tf.train.Int64List(value=values))


def float64_list_feature(values):
    """Return a TF-Feature of float64_list.

    Args:
        values: A scalar or list of values.
    Returns:
        A TF-Feature.
    """
    if not isinstance(values, collections.Iterable):
        values = [values]

    return tf.train.Feature(float_list=tf.train.FloatList(value=values))


def bytes_list_feature(values, multiple=False):
    """Return a TF-Feature of bytes.

    Args:
        values: A string.
    Returns:
        A TF-Feature.
    """
    def norm2bytes(value):
        return value.encode() if isinstance(value, str) else value

    if multiple:
        return tf.train.Feature(
            bytes_list=tf.train.BytesList(value=[norm2bytes(value)
                                                 for value in values]))
    else:
        return tf.train.Feature(
            bytes_list=tf.train.BytesList(value=[norm2bytes(values)]))



import io
import tensorflow as tf
import PIL


def iterate_tfrecord(filename):
    """Iterate through a tfrecord file.

    Args:
        filename (str): Filename to iterate.
        decode (bool): Optionally pass all records to example decoder function.
            False by default.

    Returns: Iterator of tfrecords.
    """
    for record in tf.python_io.tf_record_iterator(filename):
        example = tf.train.Example()
        example.ParseFromString(record)
        yield example


def save_tfrecords(records, output_filename):
    """Save all tfrecord examples to file.

    Args:
        records (Iterator[tf.train.Example]): Iterator of records to save.
        output_filename (str): Output file to save to.
    """
    with tf.python_io.TFRecordWriter(output_filename) as tfrecord_writer:
        for record in records:
            tfrecord_writer.write(record.SerializeToString())


def get_jpeg_string(image):
    """Builds PNG string from mask array.

    Args:
        image (PIL.Image): Mask array to generate PNG string from.

    Returns: String of mask encoded as a PNG.
    """
    # Save the new image to a PNG byte string.
    byte_buffer = io.BytesIO()
    image.save(byte_buffer, format='jpeg')
    byte_buffer.seek(0)
    return byte_buffer.read()


def get_png_string(mask_array):
    """Builds PNG string from mask array.

    Args:
        mask_array (HxW): Mask array to generate PNG string from.

    Returns: String of mask encoded as a PNG.
    """
    # Convert the new mask back to an image.
    image = PIL.Image.fromarray(mask_array.astype('uint8')).convert('RGB')
    # Save the new image to a PNG byte string.
    byte_buffer = io.BytesIO()
    image.save(byte_buffer, format='png')
    byte_buffer.seek(0)
    return byte_buffer.read()


class TFRecordHelper(object):
    """"Base class for interfacing with TFRecords. """

    def __init__(self, *filenames, output_filename=None):
        """
        Args:
            filenames: List of filenames to read from
            output_filename: Optional output_filename to use if multiple
                input filenames.
        """
        self.filenames = filenames
        self.output_filename = output_filename
        self._generator = None

    def __iter__(self):
        self._generator = self.read()
        return self

    def __next__(self):
        return next(self._generator)

    def read(self, decode=False):
        for filename in self.filenames:
            for record in iterate_tfrecord(filename):
                if decode:
                    yield self.decode_single_example(record)
                else:
                    yield record

    def write(self, records):
        if len(self.filenames) > 1 and not self.output_filename:
            raise Exception("Must provide output filename if more than one "
                            "input filename.")
        save_tfrecords(records, self.filename)

    @classmethod
    def decode_tensor(cls, example):
        """Decodes a tensor, used when creating dataset for training.

        Arguments:
            example (what type????): example tensor
        Returns: Dict[str, tf.Tensor (CHANGE)]
        """
        raise NotImplementedError()

    @classmethod
    def decode_single_example(cls, example):
        """Decode example to dictionary for easy interactive use.
        """
        raise NotImplementedError()

    @classmethod
    def build_example(cls, *args):
        """Build tfrecord example from input data.

        Used to create initial dataset.
        """
        raise NotImplementedError()
